Word2Vec model;
me.dir() + "glove-wiki-gigaword-50.txt" => string filepath;
<<< "loading model:", filepath >>>;
<<< "(this could take a few seconds)...", "" >>>;
if( !model.load( filepath ) )
{
    <<< "cannot load model:", filepath >>>;
    me.exit();
}

1 => int WORDS_PER_LINE;
4 => int LINES_PER_STANZA;
3 => int NUM_STANZAS;

// number of nearest words to retrieve for each word
// higher this number the higher the variance per word
8 => int K_NEAREST;

// word vector
float vec[model.dim()];
// used for scaling up word timing to word length
dur addTime;


// ranges for each dimension (for sound mapping)
float mins[0], maxs[0];
// get bounds for each dimension
model.minMax( mins, maxs );

// timing
400::ms => dur T_WORD; // duration per word
false => int shouldScaleTimeToWordLength; // longer words take longer?
T_WORD => dur T_LINE_PAUSE; // a little pause after each line
T_WORD * 2 => dur T_STANZA_PAUSE; // pause after each stanza

[ 31, 29, 28, 26, 24, 23, 21, 19, 17, 16, 14, 12 ] @=> int notes[];

// sound
ModalBar crystal => NRev reverb;
reverb => Gain left => dac.left;
reverb => Gain right => dac.right;
0 => crystal.preset;
.1 => reverb.mix;
// feedback delay for echos
left => DelayL DL => left; .3 => DL.gain; T_WORD*1.5 => DL.max => DL.delay;
right => DelayL DR => right; .5 => DR.gain; T_WORD*2 => DR.max => DR.delay;

// make a ConsoleInput
ConsoleInput in;
// tokenizer
StringTokenizer tok;
// line
string line[0];

// loop
while( true )
{
    // prompt
    in.prompt( "Enter a noun and an adjective to describe the noun separated by a space => (e.g. rachel amazing)" ) => now;
    
    // read
    while( in.more() )
    {
        // clear tokens array
        line.clear();
        // get the input line as one string; send to tokenizer
        tok.set( in.getLine() );
        // tokenize line into words separated by white space
        while( tok.more() )
        {
            // put into array (lowercased)
            line << tok.next().lower();
        }
         
        // if non-empty
        if( line.size() )
        {
            // do something with the input
            poem( line );
        }
    }
}

fun void poem(string input[])
{
    // similarity search results from word2vec
    string words[K_NEAREST];
    string adjectiveCandidates[K_NEAREST];
    string adjectives[0];
    input[0] => string STARTING_WORD;
    input[1] => string STARTING_ADJECTIVE;
    STARTING_WORD => string word;
    adjectives << STARTING_ADJECTIVE;
    
    model.getSimilar( word, words.size(), words );

    // line break
    chout <= IO.newline(); chout.flush();

    // loop over lines in a stanza
    for( int i; i < K_NEAREST; i++ )
    {
        if (i > 0) 
        {
            W2V.analogy( model, STARTING_WORD, STARTING_ADJECTIVE, words[i], K_NEAREST ) @=> adjectiveCandidates;
            for( int j; j < K_NEAREST; j++ ) 
            {
                0 => int adjectiveUsed;
                for( int k; k < adjectives.size(); k++ ) 
                {
                    if (adjectives[k] == adjectiveCandidates[j]) {
                        1 => adjectiveUsed;
                        break;
                    }
                }
                if (adjectiveUsed == 0)
                {
                    adjectives << adjectiveCandidates[j];
                    break;
                }
            }
        }

        say(words[i]); play(notes[i] + 40); wait();
        say("is"); play(notes[i + 2] + 40); wait();
        say(adjectives[adjectives.size() - 1]); play(notes[i + 4] + 40); wait();
        endl(); wait();
    }

    
    // print at the end
    chout <= IO.newline(); chout.flush();
    chout <= "\"Analogy Poem\"" <= IO.newline();
    chout <= IO.newline(); chout.flush();
    chout.flush();
}

// say a word with space after
fun void say( string word )
{
    say( word, " " );
}

// say a word
fun void say( string word, string append )
{
    // print it
    chout <= word <= append; chout.flush();
}

// sonify
fun void play( int pitch )
{
    play( pitch, Math.random2f(.8,1) );
}

// sonify
fun void play( int pitch, float velocity )
{
    // convert pitch to frequency and set it
    pitch => Std.mtof => crystal.freq;
    // note on
    velocity => crystal.noteOn;
}

// wait
fun void wait()
{
    wait( T_WORD );
}

// wait
fun void wait( dur T )
{    
    // let time pass, let sound...sound
    T => now;
}

// new line with timing
fun void endl()
{
    endl( T_WORD );
}

// new line with timing
fun void endl( dur T )
{
    // new line
    chout <= IO.newline(); chout.flush();
    // let time pass
    T => now;
}

//------------------------------------------------------------------------------
// word2vec helper
//------------------------------------------------------------------------------
class W2V
{
    fun static void copy( float to[], float from[] )
    {
        to.size(from.size());
        // add into element
        for( int i; i < to.size(); i++ ) from[i] => to[i];
    }
    
    fun static float[] dup( float from[] )
    {
        float to[from.size()];
        // add into element
        for( int i; i < to.size(); i++ ) from[i] => to[i];
        // return
        return to;
    }
    
    // result stored in x
    fun static void add( float x[], float y[] )
    {
        // smaller of the two
        x.size() < y.size() ? x.size() : y.size() => int size;
        // add into element
        for( int i; i < size; i++ ) y[i] +=> x[i];
    }
    
    // result stored in x
    fun static void minus( float x[], float y[] )
    {
        // smaller of the two
        x.size() < y.size() ? x.size() : y.size() => int size;
        // add into element
        for( int i; i < size; i++ ) y[i] -=> x[i];
    }
    
    fun static void scale( float x[], float scalar )
    {
        // scale
        for( int i; i < x.size(); i++ ) scalar *=> x[i];
    }
    
    fun static void scale( float result[], float x[], float scalar )
    {
        // scale
        for( int i; i < x.size(); i++ ) scalar * x[i] => result[i];
    }
    
    fun static string[] eval( Word2Vec @ w, string expr, int k )
    {
        // init 
        int pos;
        1.0 => float multiplier;
        string word;
        float wordVector[w.dim()];
        float exprVector[w.dim()];
        
        // compute
        while( true )
        {
            expr.find(" ") => pos;
            if( pos == -1 )
            {
                w.getVector(expr, wordVector);
                for( 0 => int i; i < w.dim(); i++ )
                {
                    wordVector[i] * multiplier +=> exprVector[i];
                }
                break;
            }
            else 
            {
                expr.substring(0, pos) => word;
                w.getVector(word, wordVector);
                for( 0 => int i; i < w.dim(); i++ )
                {
                    wordVector[i] * multiplier +=> exprVector[i];
                }
                expr.substring(pos + 1) => expr;
                if( expr.charAt(0) == '+' )
                {
                    1.0 => multiplier;
                    expr.substring(2) => expr;
                }
                else
                {
                    -1.0 => multiplier;
                    expr.substring(2) => expr;
                }
            }
        }
        
        // return
        string results[k];
        w.getSimilar( exprVector, k, results );
        
        return results;
    }
    
    // logical analogy: A to B is as C is to [what this function returns]
    fun static string[] analogy( Word2Vec @ w, string A, string B, string C, int k )
    {
        // the vector sum
        B + " - " + A + " + " + C => string v;
        return eval( w, v, k );
    }
}